@@ -546,6 +546,18 @@ data KnownExtension =
 
   | UnboxedSums
 
+  | TypeInType
+
+  | Strict
+
+  | StrictData
+
+  -- | Enable deriving instances via types of the same runtime representation.
+  -- Implies 'DerivingStrategies'.
+  | DerivingVia
+
+  | QuantifiedConstraints
+
   deriving (Show, Read, Eq, Ord, Enum, Bounded, Data, Typeable)
 
 -- | Certain extensions imply other extensions, and this function
@@ -581,9 +593,12 @@ impliesExts = go
                     ImpredicativeTypes   -> [ExplicitForAll]
                     PolyKinds           -> [KindSignatures]
                     TypeFamilyDependencies -> [TypeFamilies]
+                    TypeInType          -> [PolyKinds, DataKinds, KindSignatures]
+                    TypeOperators       -> [ExplicitNamespaces]
                     -- Deprecations
                     RecordPuns          -> [NamedFieldPuns]
                     PatternSignatures   -> [ScopedTypeVariables]
+                    DerivingVia         -> [DerivingStrategies]
                     _                   -> []
 
 -- | The list of extensions enabled by
