@@ -480,14 +480,17 @@ instance  Pretty (Decl l) where
         pretty (SpliceDecl _ splice) =
                 pretty splice
 
+        pretty (TSpliceDecl _ splice) =
+                pretty splice
+
         pretty (TypeSig _ nameList qualType) =
                 mySep ((punctuate comma . map pretty $ nameList)
                       ++ [text "::", pretty qualType])
 
         --  Req can be ommitted if it is empty
         --  We must print prov if req is nonempty
-        pretty (PatSynSig _ ns mtvs prov req t) =
-                let contexts = map (maybePP pretty) [prov, req]
+        pretty (PatSynSig _ ns mtvs prov mtvs2 req t) =
+                let contexts = [maybePP pretty prov, ppForall mtvs2, maybePP pretty req]
                  in
                   mySep ( [text "pattern" ]
                            ++ punctuate comma (map pretty ns)
@@ -769,12 +772,12 @@ instance  Pretty (QualConDecl l) where
                 myFsep [ppForall tvs, maybePP pretty ctxt, pretty con]
 
 instance  Pretty (GadtDecl l) where
-        pretty (GadtDecl _pos name names ty) =
+        pretty (GadtDecl _pos name tvs ctxt names ty) =
             case names of
                 Nothing ->
                     myFsep [pretty name, text "::", pretty ty]
                 Just ts' ->
-                    myFsep [pretty name, text "::" ,
+                    myFsep [pretty name, text "::" , ppForall tvs, maybePP pretty ctxt,
                          braceList . map pretty $ ts', text "->", pretty ty]
 
 instance  Pretty (ConDecl l) where
@@ -807,15 +810,31 @@ instance Pretty (Unpackedness l) where
         pretty NoUnpackPragma {} = empty
 
 instance Pretty (Deriving l) where
-  pretty (Deriving _ mds [d]) = text "deriving" <+> maybePP pretty mds <+> pretty d
-  pretty (Deriving _ mds d)   = text "deriving" <+> maybePP pretty mds <+> parenList (map pretty d)
+  pretty (Deriving _ mds d) =
+    hsep [ text "deriving"
+         , pp_strat_before
+         , pp_dct
+         , pp_strat_after ]
+    where
+      pp_dct =
+        case d of
+          [d'] -> pretty d'
+          _    -> parenList (map pretty d)
+
+      -- @via@ is unique in that in comes /after/ the class being derived,
+      -- so we must special-case it.
+      (pp_strat_before, pp_strat_after) =
+        case mds of
+          Just (via@DerivVia{}) -> (empty, pretty via)
+          _                     -> (maybePP pretty mds, empty)
 
 instance Pretty (DerivStrategy l) where
-  pretty ds = text $
+  pretty ds =
     case ds of
-      DerivStock _    -> "stock"
-      DerivAnyclass _ -> "anyclass"
-      DerivNewtype _  -> "newtype"
+      DerivStock _    -> text "stock"
+      DerivAnyclass _ -> text "anyclass"
+      DerivNewtype _  -> text "newtype"
+      DerivVia _ ty   -> text "via" <+> pretty ty
 
 ------------------------- Types -------------------------
 ppBType :: Type l -> Doc
@@ -833,6 +852,7 @@ prec_atype = 2  -- argument of type or data constructor, or of a class
 instance  Pretty (Type l) where
         prettyPrec p (TyForall _ mtvs ctxt htype) = parensIf (p > 0) $
                 myFsep [ppForall mtvs, maybePP pretty ctxt, pretty htype]
+        prettyPrec _ (TyStar _) = text "*"
         prettyPrec p (TyFun _ a b) = parensIf (p > 0) $
                 myFsep [ppBType a, text "->", pretty b]
         prettyPrec _ (TyTuple _ bxd l) =
@@ -852,7 +872,6 @@ instance  Pretty (Type l) where
         prettyPrec _ (TyVar _ name) = pretty name
         prettyPrec _ (TyCon _ name) = pretty name
         prettyPrec _ (TyParen _ t) = parens (pretty t)
---        prettyPrec _ (TyPred asst) = pretty asst
         prettyPrec _ (TyInfix _ a op b) = myFsep [pretty a, pretty op, pretty b]
         prettyPrec _ (TyKind _ t k) = parens (myFsep [pretty t, text "::", pretty k])
         prettyPrec _ (TyPromoted _ p) = pretty p
@@ -894,16 +913,6 @@ ppForall (Just vs) =    myFsep (text "forall" : map pretty vs ++ [char '.'])
 
 ---------------------------- Kinds ----------------------------
 
-instance  Pretty (Kind l) where
-        prettyPrec _ KindStar{}      = text "*"
-        prettyPrec n (KindFn _ a b)  = parensIf (n > 0) $ myFsep [prettyPrec 1 a, text "->", pretty b]
-        prettyPrec _ (KindParen _ k) = parens $ pretty k
-        prettyPrec _ (KindVar _ n)   = pretty n
-        prettyPrec _ (KindTuple _ t) = parenList . map pretty $ t
-        prettyPrec _ (KindList _ l)  = brackets .  pretty $ l
-        prettyPrec n (KindApp _ a b) =
-          parensIf (n > 3) $ myFsep [prettyPrec 3 a, prettyPrec 4 b]
-
 ppOptKind :: Maybe (Kind l) -> [Doc]
 ppOptKind Nothing  = []
 ppOptKind (Just k) = [text "::", pretty k]
@@ -1103,6 +1112,7 @@ ppLetExp l b = myFsep [text "let" <+> ppBody letIndent (ppDecls False l),
 
 instance  Pretty (Bracket l) where
         pretty (ExpBracket _ e) = ppBracket "[|" e
+        pretty (TExpBracket _ e) = myFsep [text "[||", pretty e, text "||]"]
         pretty (PatBracket _ p) = ppBracket "[p|" p
         pretty (TypeBracket _ t) = ppBracket "[t|" t
         pretty (DeclBracket _ d) =
@@ -1113,6 +1123,9 @@ ppBracket o x = myFsep [text o, pretty x, text "|]"]
 
 instance  Pretty (Splice l) where
         pretty (IdSplice _ s) = char '$' <> text s
+        pretty (TIdSplice _ s) = char '$' <> char '$' <> text s
+        pretty (TParenSplice _ e) =
+                myFsep [text "$$(", pretty e, char ')']
         pretty (ParenSplice _ e) =
                 myFsep [text "$(", pretty e, char ')']
 
@@ -1340,15 +1353,10 @@ instance (Pretty (Context l)) where
   pretty (CxSingle _ ctxt)  = pretty ctxt <+> text "=>"
   pretty (CxTuple _ context) = mySep [parenList (map pretty context), text "=>"]
 
--- hacked for multi-parameter type classes
 instance  Pretty (Asst l) where
-        pretty (ClassA _ a ts)   = myFsep $ pretty a : map ppAType ts
-        pretty (AppA _ n ns)     = myFsep $ pretty n : map pretty ns
-        pretty (InfixA _ a op b) = myFsep [pretty a, ppQNameInfix op, pretty b]
+        pretty (TypeA _ t)       = pretty t
         pretty (IParam _ i t)    = myFsep [pretty i, text "::", pretty t]
-        pretty (EqualP _ t1 t2)  = myFsep [pretty t1, text "~", pretty t2]
         pretty (ParenA _ a)      = parens (pretty a)
-        pretty (WildCardA _ mn)  = char  '_' <> maybePP pretty mn
 
 -- Pretty print a source location, useful for printing out error messages
 instance Pretty SrcLoc where
@@ -1659,17 +1667,14 @@ instance SrcInfo loc => Pretty (P.PContext loc) where
         pretty (P.CxTuple _ assts) = myFsep [parenList (map pretty assts), text "=>"]
 
 instance SrcInfo loc => Pretty (P.PAsst loc) where
-        pretty (P.ClassA _ a ts)   = myFsep $ pretty ( a) : map (prettyPrec prec_atype) ts
-        pretty (P.AppA _ n ns)     = myFsep $ pretty n : map (prettyPrec prec_atype) ns
-        pretty (P.InfixA _ a op b) = myFsep [pretty a, ppQNameInfix op, pretty b]
+        pretty (P.TypeA _ t)       = pretty t
         pretty (P.IParam _ i t)    = myFsep [pretty i, text "::", pretty t]
-        pretty (P.EqualP _ t1 t2)  = myFsep [pretty t1, text "~", pretty t2]
         pretty (P.ParenA _ a)      = parens (pretty a)
-        pretty (P.WildCardA _ mn)  = char '_' <> maybePP pretty mn
 
 instance SrcInfo loc => Pretty (P.PType loc) where
         prettyPrec p (P.TyForall _ mtvs ctxt htype) = parensIf (p > 0) $
                 myFsep [ppForall mtvs, maybePP pretty ctxt, pretty htype]
+        prettyPrec _ (P.TyStar _) = text "*"
         prettyPrec p (P.TyFun _ a b) = parensIf (p > 0) $
                 myFsep [prettyPrec prec_btype a, text "->", pretty b]
         prettyPrec _ (P.TyTuple _ bxd l) =
@@ -1693,6 +1698,7 @@ instance SrcInfo loc => Pretty (P.PType loc) where
         prettyPrec _ (P.TyInfix _ a op b) = myFsep [pretty a, pretty op, pretty b]
         prettyPrec _ (P.TyKind _ t k) = parens (myFsep [pretty t, text "::", pretty k])
         prettyPrec _ (P.TyPromoted _ p) = pretty p
+        prettyPrec _ (P.TyEquals _ a b) = myFsep [pretty a, text "~", pretty b]
         prettyPrec _ (P.TySplice _ s) = pretty s
         prettyPrec _ (P.TyBang _ b u t) = pretty u <+> pretty b <> prettyPrec prec_atype t
         prettyPrec _ (P.TyWildCard _ mn) = char '_' <> maybePP pretty mn
