@@ -65,7 +65,7 @@ module Language.Haskell.Exts.Syntax (
     -- * Class Assertions and Contexts
     Context(..), FunDep(..), Asst(..),
     -- * Types
-    Type(..), Boxed(..), Kind(..), TyVarBind(..), Promoted(..),
+    Type(..), Boxed(..), Kind, TyVarBind(..), Promoted(..),
     TypeEqn (..),
     -- * Expressions
     Exp(..), Stmt(..), QualStmt(..), FieldUpdate(..),
@@ -97,14 +97,14 @@ module Language.Haskell.Exts.Syntax (
     -- ** Main function of a program
     main_name,
     -- ** Constructors
-    unit_con_name, tuple_con_name, list_cons_name, unboxed_singleton_con_name,
+    unit_con_name, tuple_con_name, list_con_name, list_cons_name, unboxed_singleton_con_name,
     unit_con, tuple_con, unboxed_singleton_con,
     -- ** Special identifiers
     as_name, qualified_name, hiding_name, minus_name, bang_name, dot_name, star_name,
     export_name, safe_name, unsafe_name, interruptible_name, threadsafe_name,
     stdcall_name, ccall_name, cplusplus_name, dotnet_name, jvm_name, js_name,
     javascript_name, capi_name, forall_name, family_name, role_name, hole_name,
-    stock_name, anyclass_name,
+    stock_name, anyclass_name, via_name,
     -- ** Type constructors
     unit_tycon_name, fun_tycon_name, list_tycon_name, tuple_tycon_name, unboxed_singleton_tycon_name,
     unit_tycon, fun_tycon, list_tycon, tuple_tycon, unboxed_singleton_tycon,
@@ -134,7 +134,7 @@ data ModuleName l = ModuleName l String
 -- data constructors.
 data SpecialCon l
     = UnitCon l             -- ^ unit type and data constructor @()@
-    | ListCon l             -- ^ list type constructor @[]@
+    | ListCon l             -- ^ list type and data constructor @[]@
     | FunCon  l             -- ^ function type constructor @->@
     | TupleCon l Boxed Int  -- ^ /n/-ary tuple type and data
                             --   constructors @(,)@ etc, possibly boxed @(\#,\#)@
@@ -241,7 +241,7 @@ data ImportDecl l = ImportDecl
 -- | An explicit import specification list.
 data ImportSpecList l
     = ImportSpecList l Bool [ImportSpec l]
-            -- A list of import specifications.
+            -- ^ A list of import specifications.
             -- The 'Bool' is 'True' if the names are excluded
             -- by @hiding@.
   deriving (Eq,Ord,Show,Typeable,Data,Foldable,Traversable,Functor,Generic)
@@ -299,9 +299,13 @@ data Decl l
      -- ^ A declaration of default types
      | SpliceDecl   l (Exp l)
      -- ^ A Template Haskell splicing declaration
+     | TSpliceDecl  l (Exp l)
+     -- ^ A typed Template Haskell splicing declaration
      | TypeSig      l [Name l] (Type l)
      -- ^ A type signature declaration
-     | PatSynSig    l [Name l] (Maybe [TyVarBind l]) (Maybe (Context l)) (Maybe (Context l)) (Type l)
+     | PatSynSig    l [Name l] (Maybe [TyVarBind l]) (Maybe (Context l))
+                               (Maybe [TyVarBind l]) (Maybe (Context l))
+                                                     (Type l)
      -- ^ A pattern synonym signature declation
      | FunBind      l [Match l]
      -- ^ A set of function binding clauses
@@ -485,12 +489,13 @@ data Deriving l = Deriving l (Maybe (DerivStrategy l)) [InstRule l]
 
 -- | Which technique the user explicitly requested when deriving an instance.
 data DerivStrategy l
-  = DerivStock l    -- ^ GHC's \"standard\" strategy, which is to implement a
-                    --   custom instance for the data type. This only works for
-                    --   certain types that GHC knows about (e.g., 'Eq', 'Show',
-                    --   'Functor' when @-XDeriveFunctor@ is enabled, etc.)
-  | DerivAnyclass l -- ^ @-XDeriveAnyClass@
-  | DerivNewtype l  -- ^ @-XGeneralizedNewtypeDeriving@
+  = DerivStock l        -- ^ GHC's \"standard\" strategy, which is to implement a
+                        --   custom instance for the data type. This only works for
+                        --   certain types that GHC knows about (e.g., 'Eq', 'Show',
+                        --   'Functor' when @-XDeriveFunctor@ is enabled, etc.)
+  | DerivAnyclass l     -- ^ @-XDeriveAnyClass@
+  | DerivNewtype l      -- ^ @-XGeneralizedNewtypeDeriving@
+  | DerivVia l (Type l) -- ^ @-XDerivingVia@
   deriving (Eq,Ord,Show,Typeable,Data,Foldable,Traversable,Functor,Generic)
 
 -- | A binding group inside a @let@ or @where@ clause.
@@ -558,7 +563,9 @@ data FieldDecl l = FieldDecl l [Name l] (Type l)
 -- then @'Maybe' ['FieldDecl' l]@ is 'Nothing', and the whole constructor's
 -- type (such as @Int -> Bool -> Ty@) is stored in the last 'Type' field.
 data GadtDecl l
-    = GadtDecl l (Name l) (Maybe [FieldDecl l]) (Type l)
+    = GadtDecl l (Name l)
+        {-forall-} (Maybe [TyVarBind l]) {- . -} (Maybe (Context l))
+        {- => -} (Maybe [FieldDecl l]) (Type l)
   deriving (Eq,Ord,Show,Typeable,Data,Foldable,Traversable,Functor,Generic)
 
 -- | Declarations inside a class declaration.
@@ -624,6 +631,7 @@ data Type l
         (Maybe [TyVarBind l])
         (Maybe (Context l))
         (Type l)                                -- ^ qualified type
+     | TyStar  l                                -- ^ @*@, the type of types
      | TyFun   l (Type l) (Type l)              -- ^ function type
      | TyTuple l Boxed [Type l]                 -- ^ tuple type, possibly boxed
      | TyUnboxedSum l [Type l]                  -- ^ unboxed tuple type
@@ -669,15 +677,7 @@ data TyVarBind l
   deriving (Eq,Ord,Show,Typeable,Data,Foldable,Traversable,Functor,Generic)
 
 -- | An explicit kind annotation.
-data Kind l
-    = KindStar  l                    -- ^ @*@, the kind of types
-    | KindFn    l (Kind l) (Kind l)  -- ^ @->@, the kind of a type constructor
-    | KindParen l (Kind l)           -- ^ a parenthesised kind
-    | KindVar   l (QName l)          -- ^ @k@, a kind variable (-XPolyKinds)
-    | KindApp   l (Kind l) (Kind l)  -- ^ @k1 k2@
-    | KindTuple l [Kind l]           -- ^ @'(k1,k2,k3)@, a promoted tuple
-    | KindList  l (Kind l)           -- ^ @'[k1]@, a promoted list literal
-  deriving (Eq,Ord,Show,Typeable,Data,Foldable,Traversable,Functor,Generic)
+type Kind = Type
 
 
 -- | A functional dependency, given on the form
@@ -694,17 +694,10 @@ data Context l
   deriving (Eq,Ord,Show,Typeable,Data,Foldable,Traversable,Functor,Generic)
 
 -- | Class assertions.
---   In Haskell 98, the argument would be a /tyvar/, but this definition
---   allows multiple parameters, and allows them to be /type/s.
---   Also extended with support for implicit parameters and equality constraints.
 data Asst l
-        = ClassA l (QName l) [Type l]           -- ^ ordinary class assertion
-        | AppA l (Name l) [Type l]              -- ^ constraint kind assertion, @Dict :: cxt a => Dict cxt@
-        | InfixA l (Type l) (QName l) (Type l)  -- ^ class assertion where the class name is given infix
+        = TypeA l (Type l)                      -- ^ type assertion
         | IParam l (IPName l) (Type l)          -- ^ implicit parameter assertion
-        | EqualP l (Type l) (Type l)            -- ^ type equality constraint
         | ParenA l (Asst l)                     -- ^ parenthesised class assertion
-        | WildCardA l (Maybe (Name l))          -- ^ Context Wildcard
   deriving (Eq,Ord,Show,Typeable,Data,Foldable,Traversable,Functor,Generic)
 
 -- | /literal/
@@ -827,6 +820,7 @@ data XAttr l = XAttr l (XName l) (Exp l)
 -- | A template haskell bracket expression.
 data Bracket l
     = ExpBracket l (Exp l)        -- ^ expression bracket: @[| ... |]@
+    | TExpBracket l (Exp l)       -- ^ typed expression bracket: @[|| ... ||]@
     | PatBracket l (Pat l)        -- ^ pattern bracket: @[p| ... |]@
     | TypeBracket l (Type l)      -- ^ type bracket: @[t| ... |]@
     | DeclBracket l [Decl l]      -- ^ declaration bracket: @[d| ... |]@
@@ -835,7 +829,9 @@ data Bracket l
 -- | A template haskell splice expression
 data Splice l
     = IdSplice l String           -- ^ variable splice: @$var@
+    | TIdSplice l String          -- ^ typed variable splice: @$$var@
     | ParenSplice l (Exp l)       -- ^ parenthesised expression splice: @$(/exp/)@
+    | TParenSplice l (Exp l)      -- ^ parenthesised typed expression splice: @$$(/exp/)@
   deriving (Eq,Ord,Show,Typeable,Data,Foldable,Traversable,Functor,Generic)
 
 -- | The safety of a foreign function call.
@@ -1022,6 +1018,9 @@ unit_con_name l = Special l (UnitCon l)
 tuple_con_name :: l -> Boxed -> Int -> QName l
 tuple_con_name l b i = Special l (TupleCon l b (i+1))
 
+list_con_name :: l -> QName l
+list_con_name l = Special l (ListCon l)
+
 list_cons_name :: l -> QName l
 list_cons_name l = Special l (Cons l)
 
@@ -1052,7 +1051,7 @@ hole_name      l = Special l (ExprHole l)
 export_name, safe_name, unsafe_name, interruptible_name, threadsafe_name,
   stdcall_name, ccall_name, cplusplus_name, dotnet_name,
   jvm_name, js_name, javascript_name, capi_name, forall_name,
-  family_name, role_name, stock_name, anyclass_name :: l -> Name l
+  family_name, role_name, stock_name, anyclass_name, via_name :: l -> Name l
 export_name     l = Ident l "export"
 safe_name       l = Ident l "safe"
 unsafe_name     l = Ident l "unsafe"
@@ -1071,6 +1070,7 @@ family_name     l = Ident l "family"
 role_name       l = Ident l "role"
 stock_name      l = Ident l "stock"
 anyclass_name   l = Ident l "anyclass"
+via_name        l = Ident l "via"
 
 unit_tycon_name, fun_tycon_name, list_tycon_name, unboxed_singleton_tycon_name :: l -> QName l
 unit_tycon_name l = unit_con_name l
@@ -1249,7 +1249,12 @@ instance Annotated DerivStrategy where
     ann (DerivStock l)    = l
     ann (DerivAnyclass l) = l
     ann (DerivNewtype l)  = l
-    amap = fmap
+    ann (DerivVia l _)    = l
+
+    amap f (DerivStock l)    = DerivStock (f l)
+    amap f (DerivAnyclass l) = DerivAnyclass (f l)
+    amap f (DerivNewtype l)  = DerivNewtype (f l)
+    amap f (DerivVia l t)    = DerivVia (f l) t
 
 instance Annotated TypeEqn where
     ann (TypeEqn l _ _) = l
@@ -1283,8 +1288,9 @@ instance Annotated Decl where
         InfixDecl    l _ _ _            -> l
         DefaultDecl  l _                -> l
         SpliceDecl   l _                -> l
+        TSpliceDecl  l _                -> l
         TypeSig      l _ _              -> l
-        PatSynSig    l _ _ _ _ _        -> l
+        PatSynSig    l _ _ _ _ _ _      -> l
         FunBind      l _                -> l
         PatBind      l _ _ _            -> l
         ForImp       l _ _ _ _ _        -> l
@@ -1320,8 +1326,9 @@ instance Annotated Decl where
         InfixDecl    l a k ops           -> InfixDecl (f l) a k ops
         DefaultDecl  l ts                -> DefaultDecl (f l) ts
         SpliceDecl   l sp                -> SpliceDecl (f l) sp
+        TSpliceDecl  l sp                -> TSpliceDecl (f l) sp
         TypeSig      l ns t              -> TypeSig (f l) ns t
-        PatSynSig    l n dh c1 c2 t      -> PatSynSig (f l) n dh c1 c2 t
+        PatSynSig    l n dh c1 dh2 c2 t      -> PatSynSig (f l) n dh c1 dh2 c2 t
         FunBind      l ms                -> FunBind (f l) ms
         PatBind      l p rhs bs          -> PatBind (f l) p rhs bs
         ForImp       l cc msf s n t      -> ForImp (f l) cc msf s n t
@@ -1434,8 +1441,8 @@ instance Annotated FieldDecl where
     amap f (FieldDecl l ns t) = FieldDecl (f l) ns t
 
 instance Annotated GadtDecl where
-    ann (GadtDecl l _  _ _) = l
-    amap f (GadtDecl l n t1 t2) = GadtDecl (f l) n t1 t2
+    ann (GadtDecl l _ _ _ _ _) = l
+    amap f (GadtDecl l n t1 t2 t3 t4) = GadtDecl (f l) n t1 t2 t3 t4
 
 instance Annotated ClassDecl where
     ann (ClsDecl    l _)      = l
@@ -1494,6 +1501,7 @@ instance Annotated GuardedRhs where
 instance Annotated Type where
     ann t = case t of
       TyForall l _ _ _              -> l
+      TyStar  l                     -> l
       TyFun   l _ _                 -> l
       TyTuple l _ _                 -> l
       TyUnboxedSum l _              -> l
@@ -1513,6 +1521,7 @@ instance Annotated Type where
       TyQuasiQuote l _ _            -> l
     amap f t1 = case t1 of
       TyForall l mtvs mcx t         -> TyForall (f l) mtvs mcx t
+      TyStar  l                     -> TyStar (f l)
       TyFun   l t1' t2              -> TyFun (f l) t1' t2
       TyTuple l b ts                -> TyTuple (f l) b ts
       TyUnboxedSum l s              -> TyUnboxedSum (f l) s
@@ -1545,22 +1554,6 @@ instance Annotated TyVarBind where
     amap f (KindedVar   l n k) = KindedVar   (f l) n k
     amap f (UnkindedVar l n)   = UnkindedVar (f l) n
 
-instance Annotated Kind where
-    ann (KindStar l) = l
-    ann (KindFn   l _ _) = l
-    ann (KindParen l _)  = l
-    ann (KindVar l _)    = l
-    ann (KindApp l _ _)  = l
-    ann (KindTuple l _)  = l
-    ann (KindList  l _)  = l
-    amap f (KindStar l) = KindStar (f l)
-    amap f (KindFn   l k1 k2) = KindFn (f l) k1 k2
-    amap f (KindParen l k) = KindParen (f l) k
-    amap f (KindVar l n) = KindVar (f l) n
-    amap f (KindApp l k1 k2) = KindApp (f l) k1 k2
-    amap f (KindTuple l ks) = KindTuple (f l) ks
-    amap f (KindList  l ks) = KindList  (f l) ks
-
 instance Annotated FunDep where
     ann (FunDep l _ _) = l
     amap f (FunDep l ns1 ns2) = FunDep (f l) ns1 ns2
@@ -1575,21 +1568,13 @@ instance Annotated Context where
 
 instance Annotated Asst where
     ann asst = case asst of
-        ClassA l _ _     -> l
-        AppA l _ _       -> l
-        InfixA l _ _ _   -> l
+        TypeA l _        -> l
         IParam l _ _     -> l
-        EqualP l _ _     -> l
         ParenA l _       -> l
-        WildCardA l _    -> l
     amap f asst = case asst of
-        ClassA l qn ts      -> ClassA (f l) qn ts
-        AppA   l n ns       -> AppA   (f l) n ns
-        InfixA l ta qn tb   -> InfixA (f l) ta qn tb
+        TypeA l t           -> TypeA (f l) t
         IParam l ipn t      -> IParam (f l) ipn t
-        EqualP l t1 t2      -> EqualP (f l) t1 t2
         ParenA l a          -> ParenA (f l) a
-        WildCardA l mn      -> WildCardA (f l) mn
 
 instance Annotated Literal where
     ann lit = case lit of
@@ -1746,19 +1731,25 @@ instance Annotated XAttr where
 
 instance Annotated Bracket where
     ann (ExpBracket l _)  = l
+    ann (TExpBracket l _) = l
     ann (PatBracket l _)  = l
     ann (TypeBracket l _) = l
     ann (DeclBracket l _) = l
     amap f (ExpBracket l e) = ExpBracket (f l) e
+    amap f (TExpBracket l e) = TExpBracket (f l) e
     amap f (PatBracket l p) = PatBracket (f l) p
     amap f (TypeBracket l t) = TypeBracket (f l) t
     amap f (DeclBracket l ds) = DeclBracket (f l) ds
 
 instance Annotated Splice where
     ann (IdSplice l _)    = l
+    ann (TIdSplice l _)   = l
     ann (ParenSplice l _) = l
+    ann (TParenSplice l _) = l
     amap f (IdSplice l s) = IdSplice (f l) s
+    amap f (TIdSplice l s) = TIdSplice (f l) s
     amap f (ParenSplice l e) = ParenSplice (f l) e
+    amap f (TParenSplice l e) = TParenSplice (f l) e
 
 instance Annotated Safety where
     ann (PlayRisky l) = l
